/*
 *  Gideon, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "headers.h"
#include "forward.h"
#include "gideon_impl.h"

#include "views2.h"
#include "palette.h"
#include "designer.h"

extern "C" {

GIDEON_API gint gideon_version(void) {
	return GIDEON_VERSION;
}

GIDEON_API void gideon_enter_check(const gchar * msg) {
	Gtk::Main::init_gtkmm_internals();
	guiloader_enter_check(msg);
	Gideon::GideonInit::enter(msg ? msg : "");
}

GIDEON_API void gideon_leave(void) {
	Gideon::GideonInit::leave();
	guiloader_leave();
}

GIDEON_API GideonDesigner * gideon_designer_new(void) {
	Gideon::PDesigner designer= Gideon::Designer::create();
	designer->reference();
	g_object_force_floating(designer->gobj());
	return reinterpret_cast<GideonDesigner *>(designer->gobj());
}

GIDEON_API void gideon_designer_extend_ui(GideonDesigner * designer, const gchar * ui, GtkActionGroup * actions) {
	Gideon::Designer::get(designer)->extend_ui(ui, Glib::wrap(actions, true));
}

GIDEON_API void gideon_designer_apply_settings(GideonDesigner * designer, GHashTable * settings) {
	Gideon::Designer::get(designer)->apply_settings(Gideon::HashTableToSettings(settings));
}

GIDEON_API void gideon_designer_store_settings(GideonDesigner * designer, GHashTable * settings) {
	Gideon::Settings map= Gideon::HashTableToSettings(settings);
	Gideon::Designer::get(designer)->store_settings(&map);
	g_hash_table_foreach_remove(settings, reinterpret_cast<GHRFunc>(&gtk_true), NULL);
	Gideon::SettingsToHashTable(map, settings);
}

GIDEON_API gboolean gideon_designer_can_save(GideonDesigner * designer) {
	return Gideon::Designer::get(designer)->can_save();
}

GIDEON_API gboolean gideon_designer_should_save(GideonDesigner * designer) {
	return Gideon::Designer::get(designer)->should_save();
}

GIDEON_API void gideon_designer_clear(GideonDesigner * designer) {
	Gideon::Designer::get(designer)->clear();
}

GIDEON_API void gideon_designer_load(GideonDesigner * designer, const gchar * data) {
	Gideon::Designer::get(designer)->load(data);
}

GIDEON_API gchar * gideon_designer_save(GideonDesigner * designer) {
	Glib::ustring data= Gideon::Designer::get(designer)->save();
	return g_strdup(data.c_str());
}

GIDEON_API void gideon_designer_saved(GideonDesigner * designer) {
	Gideon::Designer::get(designer)->saved();
}

GIDEON_API GtkAccelGroup * gideon_designer_get_accel_group(GideonDesigner * designer) {
	return Gideon::Designer::get(designer)->get_accel_group()->gobj();
}

GIDEON_API gchar * gideon_designer_save_header(GideonDesigner * designer) {
	Glib::ustring data= Gideon::Designer::get(designer)->save_header();
	return g_strdup(data.c_str());
}

GIDEON_API void gideon_designer_set_embedded(GideonDesigner * designer, gboolean embedded) {
	Gideon::Designer::get(designer)->set_embedded(embedded);
}

}

namespace Gideon {

// Designer --------------------------------------------------------------------

Designer * Designer::get(GideonDesigner * gd) {
	Gtk::Widget * widget= Glib::wrap(GTK_WIDGET(gd));
	return dynamic_cast<Designer *>(widget);
}

PDesigner Designer::create() {
	return NewRefPtr<DesignerImpl>();
}

// GideonInit ------------------------------------------------------------------

GideonInit * GideonInit::get() {
	Check(ptr());
	return ptr();
}

void GideonInit::enter(const std::string & msg) {
	if(!ptr()) {
		ptr()= new GideonInit();
		ptr()->init();
	} else
		++ptr()->refcounter;
	ptr()->snapshot(msg);
}

void GideonInit::leave() {
	Check(ptr() && ptr()->refcounter > 0);
	ptr()->check();
	--ptr()->refcounter;
	if(ptr()->refcounter == 0) {
		delete ptr();
		ptr()= NULL;
	}
}

GideonInit * & GideonInit::ptr() {
	static GideonInit * gideonInit= NULL;
	return gideonInit;
}

GideonInit::GideonInit() {
	Check(GTK_MAJOR_VERSION*1000+GTK_MINOR_VERSION <= gtk_major_version*1000+gtk_minor_version);
	Check(GUILOADER_VERSION/1000 <= guiloader_version()/1000);
	refcounter= 1;
	palette= new Palette();
}

GideonInit::~GideonInit() {
	delete palette;
}

void GideonInit::init() {
	palette->init();
}

void GideonInit::snapshot(const std::string & msg) {
	labels.push_back(msg);
	if(!msg.empty())
		snapshots.push_back(objects);
}

void GideonInit::check() {

	std::string msg= labels.back();
	labels.pop_back();

	if(!msg.empty()) {

		#ifdef GIDEON_SANDBOX
	
			const Set & snapshot= snapshots.back();
	
			Check(std::includes(objects.begin(), objects.end(), snapshot.begin(), snapshot.end()));
	
			std::list<Gideon::Object *> leaked;
			std::set_difference(
				objects.begin(), objects.end(),
				snapshot.begin(), snapshot.end(),
				std::back_inserter(leaked)
			);
	
			int leaks= leaked.size();
	
			if(leaks > 0) {
	
				std::cerr << "[Possible leaks in " << msg << ": " << ToString(leaks) << "]" << std::endl;
	
				while(!leaked.empty()) {
					Gideon::Object * obj= leaked.front();
					std::string cname= GetTypeName(typeid(*obj));
					std::cerr << "---------" << cname << ", " << obj->refcounter;
	
					CAny * any= dynamic_cast<CAny *>(obj);
					if(any && any->getType()!=InvalidTypeId) {
						std::cerr << ", "
							<< *palette->getEntry(any->getType())->getTypeName()
							<< "=" << any->toString();
					}
	
					Node * node= dynamic_cast<Node *>(obj);
					if(node) {
						std::cerr 
							<< ", name=" << *node->getName()
							<< ", out=" << ((node->getRole()==nrScalar || node->getRole()==nrLink) ? 0 : node->getOut()->size())
							<< ", in=" << node->getIn()->size();
					}
	
					std::cerr << std::endl;
					leaked.pop_front();
				}
				
			}
			
		#else
		
			int snapshot= snapshots.back();
			int leaks= objects - snapshot;
			Check(leaks >= 0);
		
		#endif
	
		if(leaks != 0) {
			Glib::ustring str= "Possible leaks in " + msg + ": " + ToString(leaks);
			#ifndef GIDEON_SANDBOX
				std::cerr << str << std::endl;
			#endif
			ShowMessage(str);
		}
	
		snapshots.pop_back();

	}

}

void GideonInit::objectCreated(Gideon::Object * object) {
	#ifdef GIDEON_SANDBOX
		Check(objects.insert(object).second);
	#else
		++objects;
	#endif
}

void GideonInit::objectDeleted(Gideon::Object * object) {
	#ifdef GIDEON_SANDBOX
		Check(objects.erase(object)==1);
	#else
		--objects;
	#endif
}

}
